@@ -35,6 +35,7 @@
 */
 
 #include "pub_core_basics.h"
+#include "pub_core_vki.h"
 #include "pub_core_threadstate.h"
 #include "pub_core_debuginfo.h"   /* self */
 #include "pub_core_demangle.h"
@@ -45,10 +46,19 @@
 #include "pub_core_options.h"
 #include "pub_core_redir.h"       // VG_(redir_notify_{new,delete}_SegInfo)
 #include "pub_core_aspacemgr.h"
+#include "pub_core_machine.h"     // VG_PLAT_USES_PPCTOC
+#include "pub_core_xarray.h"
 #include "priv_storage.h"
 #include "priv_readdwarf.h"
 #include "priv_readstabs.h"
-#include "priv_readelf.h"
+#if defined(VGO_linux)
+# include "priv_readelf.h"
+#elif defined(VGO_aix5)
+# include "pub_core_debuglog.h"
+# include "pub_core_libcproc.h"
+# include "pub_core_libcfile.h"
+# include "priv_readxcoff.h"
+#endif
 
 
 /*------------------------------------------------------------*/
@@ -76,7 +86,8 @@
    This terminology is not used consistently, but a start has been
    made.  readelf.c and the call-frame info reader in readdwarf.c now
    use it.  Specifically, various variables and structure fields have
-   been annotated with _avma / _svma / _image / _bias.
+   been annotated with _avma / _svma / _image / _bias.  In places _img
+   is used instead of _image for the sake of brevity.
 */
 
 
@@ -100,16 +111,37 @@ static SegInfo* segInfo_list = NULL;
 /* Allocate and zero out a new SegInfo record. */
 static 
 SegInfo* alloc_SegInfo(Addr start, SizeT size, OffT foffset, 
-                       const HChar* filename)
+                       const UChar* filename,
+                       const UChar* memname)
 {
-   SegInfo* si = VG_(arena_calloc)(VG_AR_SYMTAB, 1, sizeof(SegInfo));
+   Bool     traceme;
+   SegInfo* si;
 
+   vg_assert(filename);
+
+   si = VG_(arena_calloc)(VG_AR_SYMTAB, 1, sizeof(SegInfo));
    si->text_start_avma = start;
    si->text_size       = size;
    si->foffset         = foffset;
    si->filename        = VG_(arena_strdup)(VG_AR_SYMTAB, filename);
+   si->memname         = memname 
+                           ?  VG_(arena_strdup)(VG_AR_SYMTAB, memname)
+                           :  NULL;
+
+   /* Everything else -- pointers, sizes, arrays -- is zeroed by calloc.
+      Now set up the debugging-output flags. */
+   traceme 
+      = VG_(string_match)( VG_(clo_trace_symtab_patt), filename )
+        || (memname && VG_(string_match)( VG_(clo_trace_symtab_patt), 
+                                          memname ));
+   if (traceme) {
+      si->trace_symtab = VG_(clo_trace_symtab);
+      si->trace_cfi    = VG_(clo_trace_cfi);
+      si->ddump_syms   = VG_(clo_debug_dump_syms);
+      si->ddump_line   = VG_(clo_debug_dump_line);
+      si->ddump_frames = VG_(clo_debug_dump_frames);
+   }
 
-   // Everything else -- pointers, sizes, arrays -- is zeroed by calloc.
    return si;
 }
 
@@ -119,10 +151,11 @@ static void free_SegInfo ( SegInfo* si )
 {
    struct strchunk *chunk, *next;
    vg_assert(si != NULL);
-   if (si->filename) VG_(arena_free)(VG_AR_SYMTAB, si->filename);
-   if (si->symtab)   VG_(arena_free)(VG_AR_SYMTAB, si->symtab);
-   if (si->loctab)   VG_(arena_free)(VG_AR_SYMTAB, si->loctab);
-   if (si->cfsi)     VG_(arena_free)(VG_AR_SYMTAB, si->cfsi);
+   if (si->filename)   VG_(arena_free)(VG_AR_SYMTAB, si->filename);
+   if (si->symtab)     VG_(arena_free)(VG_AR_SYMTAB, si->symtab);
+   if (si->loctab)     VG_(arena_free)(VG_AR_SYMTAB, si->loctab);
+   if (si->cfsi)       VG_(arena_free)(VG_AR_SYMTAB, si->cfsi);
+   if (si->cfsi_exprs) VG_(deleteXA)(si->cfsi_exprs);
 
    for (chunk = si->strchunks; chunk != NULL; chunk = next) {
       next = chunk->next;
@@ -138,6 +171,14 @@ static void free_SegInfo ( SegInfo* si )
 */
 static void discard_SegInfo ( SegInfo* si )
 {
+#  if defined(VGP_ppc32_aix5)
+   HChar* reason = "__unload";
+#  elif defined(VGP_ppc64_aix5)
+   HChar* reason = "kunload64";
+#  else
+   HChar* reason = "munmap";
+#  endif
+
    SegInfo** prev_next_ptr = &segInfo_list;
    SegInfo*  curr          =  segInfo_list;
 
@@ -146,10 +187,11 @@ static void discard_SegInfo ( SegInfo* si )
          // Found it;  remove from list and free it.
          if (VG_(clo_verbosity) > 1 || VG_(clo_trace_redir))
             VG_(message)(Vg_DebugMsg, 
-                         "Discarding syms at %p-%p in %s due to munmap()", 
+                         "Discarding syms at %p-%p in %s due to %s()", 
                          si->text_start_avma, 
                          si->text_start_avma + si->text_size,
-                         curr->filename ? curr->filename : (UChar*)"???");
+                         curr->filename ? curr->filename : (UChar*)"???",
+                         reason);
          vg_assert(*prev_next_ptr == curr);
          *prev_next_ptr = curr->next;
          VG_(redir_notify_delete_SegInfo)( curr );
@@ -199,12 +241,29 @@ static void discard_syms_in_range ( Addr start, SizeT length )
 /* Create a new SegInfo with the specific address/length/vma offset,
    then snarf whatever info we can from the given filename into it. */
 static
-SegInfo* acquire_syms_for_range( Addr seg_addr, SizeT seg_len,
-                                 OffT seg_offset, const Char* seg_filename)
+SegInfo* acquire_syms_for_range( 
+            /* ALL        */ Addr  seg_addr, 
+            /* ALL        */ SizeT seg_len,
+            /* ELF only   */ OffT  seg_offset, 
+            /* ALL        */ const UChar* seg_filename,
+            /* XCOFF only */ const UChar* seg_memname,
+	    /* XCOFF only */ Addr  data_addr,
+	    /* XCOFF only */ SizeT data_len,
+	    /* XCOFF only */ Bool  is_mainexe
+         )
 {
-   SegInfo* si = alloc_SegInfo(seg_addr, seg_len, seg_offset, seg_filename);
+   Bool     ok;
+   SegInfo* si = alloc_SegInfo(seg_addr, seg_len, seg_offset, 
+                               seg_filename, seg_memname);
+#  if defined(VGO_linux)
+   ok = ML_(read_elf_debug_info) ( si );
+#  elif defined(VGO_aix5)
+   ok = ML_(read_xcoff_debug_info) ( si, data_addr, data_len, is_mainexe );
+#  else
+#    error Unknown OS
+#  endif
 
-   if (! ML_(read_elf_debug_info) ( si )) {
+   if (!ok) {
       // Something went wrong (eg. bad ELF file).
       free_SegInfo( si );
       si = NULL;
@@ -224,11 +283,13 @@ SegInfo* acquire_syms_for_range( Addr seg_addr, SizeT seg_len,
 }
 
 
-/*------------------------------------------------------------*/
-/*---                                                      ---*/
-/*--- TOP LEVEL: NOTIFICATION (ACQUIRE/DISCARD INFO)       ---*/
-/*---                                                      ---*/
-/*------------------------------------------------------------*/
+/*--------------------------------------------------------------*/
+/*---                                                        ---*/
+/*--- TOP LEVEL: NOTIFICATION (ACQUIRE/DISCARD INFO) (LINUX) ---*/
+/*---                                                        ---*/
+/*--------------------------------------------------------------*/
+
+#if defined(VGO_linux)
 
 /* The debug info system is driven by notifications that a text
    segment has been mapped in, or unmapped.  When that happens it
@@ -245,7 +306,7 @@ SegInfo* acquire_syms_for_range( Addr seg_addr, SizeT seg_len,
 
 void VG_(di_notify_mmap)( Addr a, Bool allow_SkFileV )
 {
-   NSegment* seg;
+   NSegment const * seg;
    HChar*    filename;
    Bool      ok;
 
@@ -289,7 +350,7 @@ void VG_(di_notify_mmap)( Addr a, Bool allow_SkFileV )
    seg = VG_(am_find_nsegment)(a);
    vg_assert(seg);
 
-   filename = VG_(am_get_filename)( seg );
+   filename = VG_(am_get_filename)( (NSegment*)seg );
    if (!filename)
       return;
 
@@ -313,7 +374,8 @@ void VG_(di_notify_mmap)( Addr a, Bool allow_SkFileV )
 
    /* .. and acquire new info. */
    acquire_syms_for_range( seg->start, seg->end + 1 - seg->start, 
-                           seg->offset, filename );
+                           seg->offset, filename,
+                           /* XCOFF only */ NULL, 0, 0, False );
 
    /* acquire_syms_for_range makes its own copy of filename, so is
       safe to free it. */
@@ -343,6 +405,71 @@ void VG_(di_notify_mprotect)( Addr a, SizeT len, UInt prot )
       discard_syms_in_range(a, len);
 }
 
+#endif /* defined(VGO_linux) */
+
+
+/*-------------------------------------------------------------*/
+/*---                                                       ---*/
+/*--- TOP LEVEL: NOTIFICATION (ACQUIRE/DISCARD INFO) (AIX5) ---*/
+/*---                                                       ---*/
+/*-------------------------------------------------------------*/
+
+#if defined(VGO_aix5)
+
+/* The supplied parameters describe a code segment and its associated
+   data segment, that have recently been mapped in -- so we need to
+   read debug info for it -- or conversely, have recently been dumped,
+   in which case the relevant debug info has to be unloaded. */
+
+void VG_(di_aix5_notify_segchange)( 
+               Addr   code_start,
+               Word   code_len,
+               Addr   data_start,
+               Word   data_len,
+               UChar* file_name,
+               UChar* mem_name,
+               Bool   is_mainexe,
+               Bool   acquire )
+{
+   SegInfo* si;
+
+   if (acquire) {
+
+      acquire_syms_for_range(
+         /* ALL        */ code_start, 
+         /* ALL        */ code_len,
+         /* ELF only   */ 0,
+         /* ALL        */ file_name,
+         /* XCOFF only */ mem_name,
+         /* XCOFF only */ data_start,
+         /* XCOFF only */ data_len,
+         /* XCOFF only */ is_mainexe 
+      );
+
+   } else {
+
+      /* Dump all the segInfos whose text segments intersect
+         code_start/code_len. */
+      while (True) {
+         for (si = segInfo_list; si; si = si->next) {
+            if (code_start + code_len <= si->text_start_avma
+                || si->text_start_avma + si->text_size <= code_start)
+               continue; /* no overlap */
+            else 
+               break;
+         }
+         if (si == NULL)
+            break;
+         /* Need to delete 'si' */
+         discard_SegInfo(si);
+      }
+
+   }
+}
+        
+
+#endif /* defined(VGO_aix5) */
+
 
 /*------------------------------------------------------------*/
 /*---                                                      ---*/
@@ -530,12 +657,26 @@ Bool VG_(get_fnname_Z_demangle_only) ( Addr a, Char* buf, Int nbuf )
    Caller supplies buf and nbuf. */
 Bool VG_(get_objname) ( Addr a, Char* buf, Int nbuf )
 {
+   Int used;
    SegInfo* si;
 
+   vg_assert(nbuf > 0);
    for (si = segInfo_list; si != NULL; si = si->next) {
       if (si->text_start_avma <= a 
           && a < si->text_start_avma + si->text_size) {
          VG_(strncpy_safely)(buf, si->filename, nbuf);
+         if (si->memname) {
+            used = VG_(strlen)(buf);
+            if (used < nbuf) 
+               VG_(strncpy_safely)(&buf[used], "(", nbuf-used);
+            used = VG_(strlen)(buf);
+            if (used < nbuf) 
+               VG_(strncpy_safely)(&buf[used], si->memname, nbuf-used);
+            used = VG_(strlen)(buf);
+            if (used < nbuf) 
+               VG_(strncpy_safely)(&buf[used], ")", nbuf-used);
+         }
+         buf[nbuf-1] = 0;
          return True;
       }
    }
@@ -599,8 +740,14 @@ Bool VG_(get_filename_linenum) ( Addr a,
               (dirname != NULL && dirname_available != NULL) );
 
    search_all_loctabs ( a, &si, &locno );
-   if (si == NULL) 
+   if (si == NULL) {
+      if (dirname_available) {
+         *dirname_available = False;
+         *dirname = 0;
+      }
       return False;
+   }
+
    VG_(strncpy_safely)(filename, si->loctab[locno].filename, n_filename);
    *lineno = si->loctab[locno].lineno;
 
@@ -623,6 +770,43 @@ Bool VG_(get_filename_linenum) ( Addr a,
 }
 
 
+/* Map a function name to its entry point and toc pointer.  Is done by
+   sequential search of all symbol tables, so is very slow.  To
+   mitigate the worst performance effects, you may specify a soname
+   pattern, and only objects matching that pattern are searched.
+   Therefore specify "*" to search all the objects.  On TOC-afflicted
+   platforms, a symbol is deemed to be found only if it has a nonzero
+   TOC pointer.  */
+Bool VG_(lookup_symbol_SLOW)(UChar* sopatt, UChar* name, Addr* pEnt, Addr* pToc)
+{
+   Bool     require_pToc = False;
+   Int      i;
+   SegInfo* si;
+   Bool     debug = False;
+#  if defined(VG_PLAT_USES_PPCTOC)
+   require_pToc = True;
+#  endif
+   for (si = segInfo_list; si; si = si->next) {
+      if (debug)
+         VG_(printf)("lookup_symbol_SLOW: considering %s\n", si->soname);
+      if (!VG_(string_match)(sopatt, si->soname)) {
+         if (debug)
+            VG_(printf)(" ... skip\n");
+         continue;
+      }
+      for (i = 0; i < si->symtab_used; i++) {
+         if (0==VG_(strcmp)(name, si->symtab[i].name)
+             && (require_pToc ? si->symtab[i].tocptr : True)) {
+            *pEnt = si->symtab[i].addr;
+            *pToc = si->symtab[i].tocptr;
+            return True;
+         }
+      }
+   }
+   return False;
+}
+
+
 /* VG_(describe_IP): print into buf info on code address, function
    name and filename. */
 
@@ -700,7 +884,9 @@ Char* VG_(describe_IP)(Addr eip, Char* buf, Int n_buf)
    static UChar buf_srcloc[BUF_LEN];
    static UChar buf_dirname[BUF_LEN];
    Bool  know_dirinfo = False;
-   Bool  know_fnname  = VG_(get_fnname) (eip, buf_fn,  BUF_LEN);
+   Bool  know_fnname  = VG_(clo_sym_offsets)
+                        ? VG_(get_fnname_w_offset) (eip, buf_fn, BUF_LEN)
+                        : VG_(get_fnname) (eip, buf_fn, BUF_LEN);
    Bool  know_objname = VG_(get_objname)(eip, buf_obj, BUF_LEN);
    Bool  know_srcloc  = VG_(get_filename_linenum)(
                            eip, 
@@ -721,7 +907,7 @@ Char* VG_(describe_IP)(Addr eip, Char* buf, Int n_buf)
          1/10 to the file name, leaving 1/10 for all the fixed-length
          stuff. */
       APPEND("<frame>");
-      VG_(sprintf)(ibuf,"<ip>0x%llx</ip>", (ULong)eip);
+      VG_(sprintf)(ibuf,"<ip>0x%llX</ip>", (ULong)eip);
       APPEND(maybe_newline);
       APPEND(ibuf);
       if (know_objname) {
@@ -759,7 +945,7 @@ Char* VG_(describe_IP)(Addr eip, Char* buf, Int n_buf)
    } else {
 
       /* Print for humans to read */
-      VG_(sprintf)(ibuf,"0x%llx: ", (ULong)eip);
+      VG_(sprintf)(ibuf,"0x%llX: ", (ULong)eip);
       APPEND(ibuf);
       if (know_fnname) { 
          APPEND(buf_fn);
@@ -792,6 +978,85 @@ Char* VG_(describe_IP)(Addr eip, Char* buf, Int n_buf)
 #  undef BUF_LEN
 }
 
+
+/*------------------------------------------------------------*/
+/*--- For unwinding the stack using                       --- */
+/*--- pre-summarised DWARF3 .eh_frame info                 ---*/
+/*------------------------------------------------------------*/
+
+/* Gather up all the constant pieces of info needed to evaluate
+   a CfiExpr into one convenient struct. */
+typedef
+   struct {
+      Addr ipHere;
+      Addr spHere;
+      Addr fpHere;
+      Addr min_accessible;
+      Addr max_accessible;
+   }
+   CfiExprEvalContext;
+
+/* Evaluate the CfiExpr rooted at ix in exprs given the context eec.
+   *ok is set to False on failure, but not to True on success.  The
+   caller must set it to True before calling. */
+static 
+UWord evalCfiExpr ( XArray* exprs, Int ix, 
+                    CfiExprEvalContext* eec, Bool* ok )
+{
+   UWord wL, wR;
+   Addr  a;
+   CfiExpr* e = VG_(indexXA)( exprs, ix );
+   switch (e->tag) {
+      case Cex_Binop:
+         wL = evalCfiExpr( exprs, e->Cex.Binop.ixL, eec, ok );
+         if (!(*ok)) return 0;
+         wR = evalCfiExpr( exprs, e->Cex.Binop.ixR, eec, ok );
+         if (!(*ok)) return 0;
+         switch (e->Cex.Binop.op) {
+            case Cop_Add: return wL + wR;
+            case Cop_Sub: return wL - wR;
+            case Cop_And: return wL & wR;
+            case Cop_Mul: return wL * wR;
+            default: goto unhandled;
+         }
+         /*NOTREACHED*/
+      case Cex_CfiReg:
+         switch (e->Cex.CfiReg.reg) {
+            case Creg_IP: return (Addr)eec->ipHere;
+            case Creg_SP: return (Addr)eec->spHere;
+            case Creg_FP: return (Addr)eec->fpHere;
+            default: goto unhandled;
+         }
+         /*NOTREACHED*/
+      case Cex_Const:
+         return e->Cex.Const.con;
+      case Cex_Deref:
+         a = evalCfiExpr( exprs, e->Cex.Deref.ixAddr, eec, ok );
+         if (!(*ok)) return 0;
+         if (a < eec->min_accessible
+             || (a + sizeof(UWord) - 1) > eec->max_accessible) {
+            *ok = False;
+            return 0;
+         }
+         /* let's hope it doesn't trap! */
+         return * ((UWord*)a);
+      default: 
+         goto unhandled;
+   }
+   /*NOTREACHED*/
+  unhandled:
+   VG_(printf)("\n\nevalCfiExpr: unhandled\n");
+   ML_(ppCfiExpr)( exprs, ix );
+   VG_(printf)("\n");
+   vg_assert(0);
+   /*NOTREACHED*/
+   return 0;
+}
+
+
+/* The main function for DWARF2/3 CFI-based stack unwinding.
+   Given an IP/SP/FP triple, produce the IP/SP/FP values for the
+   previous frame, if possible. */
 /* Returns True if OK.  If not OK, *{ip,sp,fp}P are not changed. */
 /* NOTE: this function may rearrange the order of entries in the
    SegInfo list. */
@@ -801,11 +1066,14 @@ Bool VG_(use_CF_info) ( /*MOD*/Addr* ipP,
                         Addr min_accessible,
                         Addr max_accessible )
 {
+   Bool     ok;
    Int      i;
    SegInfo* si;
    DiCfSI*  cfsi = NULL;
    Addr     cfa, ipHere, spHere, fpHere, ipPrev, spPrev, fpPrev;
 
+   CfiExprEvalContext eec;
+
    static UInt n_search = 0;
    static UInt n_steps = 0;
    n_search++;
@@ -837,7 +1105,7 @@ Bool VG_(use_CF_info) ( /*MOD*/Addr* ipP,
       VG_(printf)("%u %u\n", n_search, n_steps);
 
    /* Start of performance-enhancing hack: once every 16 (chosen
-      hackily after profiling) successful searchs, move the found
+      hackily after profiling) successful searches, move the found
       SegInfo one step closer to the start of the list.  This makes
       future searches cheaper.  For starting konqueror on amd64, this
       in fact reduces the total amount of searching done by the above
@@ -869,7 +1137,7 @@ Bool VG_(use_CF_info) ( /*MOD*/Addr* ipP,
 
    if (0) {
       VG_(printf)("found cfisi: "); 
-      ML_(ppDiCfSI)(cfsi);
+      ML_(ppDiCfSI)(si->cfsi_exprs, cfsi);
    }
 
    ipPrev = spPrev = fpPrev = 0;
@@ -878,7 +1146,36 @@ Bool VG_(use_CF_info) ( /*MOD*/Addr* ipP,
    spHere = *spP;
    fpHere = *fpP;
 
-   cfa = cfsi->cfa_off + (cfsi->cfa_sprel ? spHere : fpHere);
+   /* First compute the CFA. */
+   cfa = 0;
+   switch (cfsi->cfa_how) {
+      case CFIC_SPREL: 
+         cfa = cfsi->cfa_off + spHere;
+         break;
+      case CFIC_FPREL: 
+         cfa = cfsi->cfa_off + fpHere;
+         break;
+      case CFIC_EXPR: 
+         if (0) {
+            VG_(printf)("CFIC_EXPR: ");
+            ML_(ppCfiExpr)(si->cfsi_exprs, cfsi->cfa_off);
+            VG_(printf)("\n");
+         }
+         eec.ipHere = ipHere;
+         eec.spHere = spHere;
+         eec.fpHere = fpHere;
+         eec.min_accessible = min_accessible;
+         eec.max_accessible = max_accessible;
+         ok = True;
+         cfa = evalCfiExpr(si->cfsi_exprs, cfsi->cfa_off, &eec, &ok );
+         if (!ok) return False;
+         break;
+      default: 
+         vg_assert(0);
+   }
+
+   /* Now we know the CFA, use it to roll back the registers we're
+      interested in. */
 
 #  define COMPUTE(_prev, _here, _how, _off)             \
       do {                                              \
@@ -898,6 +1195,20 @@ Bool VG_(use_CF_info) ( /*MOD*/Addr* ipP,
             case CFIR_CFAREL:                           \
                _prev = cfa + (Word)_off;                \
                break;                                   \
+            case CFIR_EXPR:                             \
+               if (0)                                   \
+                  ML_(ppCfiExpr)(si->cfsi_exprs,_off);  \
+               eec.ipHere = ipHere;                     \
+               eec.spHere = spHere;                     \
+               eec.fpHere = fpHere;                     \
+               eec.min_accessible = min_accessible;     \
+               eec.max_accessible = max_accessible;     \
+               ok = True;                               \
+               _prev = evalCfiExpr(si->cfsi_exprs, _off, &eec, &ok ); \
+               if (!ok) return False;                   \
+               break;                                   \
+            default:                                    \
+               vg_assert(0);                            \
          }                                              \
       } while (0)
 
@@ -955,7 +1266,7 @@ VgSectKind VG_(seginfo_sect_kind)(Addr a)
    SegInfo* si;
    VgSectKind ret = Vg_SectUnknown;
 
-   for(si = segInfo_list; si != NULL; si = si->next) {
+   for (si = segInfo_list; si != NULL; si = si->next) {
       if (a >= si->text_start_avma 
           && a < si->text_start_avma + si->text_size) {
 
@@ -987,6 +1298,33 @@ VgSectKind VG_(seginfo_sect_kind)(Addr a)
    return ret;
 }
 
+Char* VG_(seginfo_sect_kind_name)(Addr a, Char* buf, UInt n_buf)
+{
+   switch (VG_(seginfo_sect_kind)(a)) {
+      case Vg_SectUnknown:
+         VG_(snprintf)(buf, n_buf, "Unknown");
+         break;
+      case Vg_SectText:
+         VG_(snprintf)(buf, n_buf, "Text");
+         break;
+      case Vg_SectData:
+         VG_(snprintf)(buf, n_buf, "Data");
+         break;
+      case Vg_SectBSS:
+         VG_(snprintf)(buf, n_buf, "BSS");
+         break;
+      case Vg_SectGOT:
+         VG_(snprintf)(buf, n_buf, "GOT");
+         break;
+      case Vg_SectPLT:
+         VG_(snprintf)(buf, n_buf, "PLT");
+         break;
+      default:
+         vg_assert(0);
+   }
+   return buf;
+}
+
 Int VG_(seginfo_syms_howmany) ( const SegInfo *si )
 {
    return si->symtab_used;
@@ -995,13 +1333,15 @@ Int VG_(seginfo_syms_howmany) ( const SegInfo *si )
 void VG_(seginfo_syms_getidx) ( const SegInfo *si, 
                                       Int idx,
                                /*OUT*/Addr*   addr,
+                               /*OUT*/Addr*   tocptr,
                                /*OUT*/UInt*   size,
                                /*OUT*/HChar** name )
 {
    vg_assert(idx >= 0 && idx < si->symtab_used);
-   if (addr) *addr = si->symtab[idx].addr;
-   if (size) *size = si->symtab[idx].size;
-   if (name) *name = (HChar*)si->symtab[idx].name;
+   if (addr)   *addr   = si->symtab[idx].addr;
+   if (tocptr) *tocptr = si->symtab[idx].tocptr;
+   if (size)   *size   = si->symtab[idx].size;
+   if (name)   *name   = (HChar*)si->symtab[idx].name;
 }
 
 
