@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2006 Julian Seward 
+   Copyright (C) 2000-2007 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -29,6 +29,7 @@
 */
 
 #include "pub_core_basics.h"
+#include "pub_core_vki.h"
 #include "pub_core_threadstate.h"
 #include "pub_core_libcassert.h"
 #include "pub_core_libcbase.h"
@@ -214,6 +215,13 @@ Bool VG_(thread_stack_next)(ThreadId* tid, Addr* stack_min, Addr* stack_max)
    return False;
 }
 
+Addr VG_(thread_get_stack_max)(ThreadId tid)
+{
+   vg_assert(0 <= tid && tid < VG_N_THREADS && tid != VG_INVALID_THREADID);
+   vg_assert(VG_(threads)[tid].status != VgTs_Empty);
+   return VG_(threads)[tid].client_stack_highest_word;
+}
+
 //-------------------------------------------------------------
 /* Details about the capabilities of the underlying (host) CPU.  These
    details are acquired by (1) enquiring with the CPU at startup, or
@@ -369,7 +377,7 @@ Bool VG_(machine_get_hwcaps)( void )
      if (__builtin_setjmp(env_sigill)) {
         have_F = False;
      } else {
-        __asm__ __volatile__("fmr 0,0");
+        __asm__ __volatile__(".long 0xFC000090"); /*fmr 0,0 */
      }
 
      /* Altivec insns */
@@ -396,7 +404,7 @@ Bool VG_(machine_get_hwcaps)( void )
      if (__builtin_setjmp(env_sigill)) {
         have_FX = False;
      } else {
-        __asm__ __volatile__("fsqrt 0,0");
+        __asm__ __volatile__(".long 0xFC00002C"); /*fsqrt 0,0 */
      }
 
      /* Graphics optional (stfiwx, fres, frsqrte, fsel) */
@@ -407,7 +415,7 @@ Bool VG_(machine_get_hwcaps)( void )
      if (__builtin_setjmp(env_sigill)) {
         have_GX = False;
      } else {
-        __asm__ __volatile__("frsqrte 0,0");
+        __asm__ __volatile__(".long 0xFC000034"); /* frsqrte 0,0 */
      }
 
      r = VG_(sigaction)(VKI_SIGILL, &saved_act, NULL);
@@ -481,7 +489,7 @@ Bool VG_(machine_get_hwcaps)( void )
      if (__builtin_setjmp(env_sigill)) {
         have_V = False;
      } else {
-        __asm__ __volatile__("vor 0,0,0");
+        __asm__ __volatile__(".long 0x10000484"); /*vor 0,0,0*/
      }
 
      /* General-Purpose optional (fsqrt, fsqrts) */
@@ -491,7 +499,7 @@ Bool VG_(machine_get_hwcaps)( void )
      if (__builtin_setjmp(env_sigill)) {
         have_FX = False;
      } else {
-        __asm__ __volatile__("fsqrt 0,0");
+        __asm__ __volatile__(".long 0xFC00002C"); /*fsqrt 0,0*/
      }
 
      /* Graphics optional (stfiwx, fres, frsqrte, fsel) */
@@ -501,7 +509,7 @@ Bool VG_(machine_get_hwcaps)( void )
      if (__builtin_setjmp(env_sigill)) {
         have_GX = False;
      } else {
-        __asm__ __volatile__("frsqrte 0,0");
+        __asm__ __volatile__(".long 0xFC000034"); /*frsqrte 0,0*/
      }
 
      VG_(sigaction)(VKI_SIGILL, &saved_act, NULL);
@@ -581,22 +589,18 @@ void VG_(machine_get_VexArchInfo)( /*OUT*/VexArch* pVa,
 
 
 // Given a pointer to a function as obtained by "& functionname" in C,
-// produce a pointer to the actual entry point for the function.  For
-// most platforms it's the identity function.  Unfortunately, on
-// ppc64-linux it isn't (sigh).
+// produce a pointer to the actual entry point for the function.
 void* VG_(fnptr_to_fnentry)( void* f )
 {
-#if defined(VGP_x86_linux)
+#if defined(VGP_x86_linux) || defined(VGP_amd64_linux) \
+                           || defined(VGP_ppc32_linux)
    return f;
-#elif defined(VGP_amd64_linux)
-   return f;
-#elif defined(VGP_ppc32_linux)
-   return f;
-#elif defined(VGP_ppc64_linux)
-   /* f is a pointer to a 3-word function descriptor, of which
-      the first word is the entry address. */
-   /* Don't ask me.  Really.  I have no idea why. */
-   ULong* descr = (ULong*)f;
+#elif defined(VGP_ppc64_linux) || defined(VGP_ppc32_aix5) \
+                               || defined(VGP_ppc64_aix5)
+   /* All other ppc variants use the AIX scheme, in which f is a
+      pointer to a 3-word function descriptor, of which the first word
+      is the entry address. */
+   UWord* descr = (UWord*)f;
    return (void*)(descr[0]);
 #else
 #  error "Unknown platform"
