@@ -7,8 +7,8 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2005 Nicholas Nethercote <njn@valgrind.org>
-   Copyright (C) 2005 Cerion Armour-Brown <cerion@open-works.co.uk>
+   Copyright (C) 2005-2007 Nicholas Nethercote <njn@valgrind.org>
+   Copyright (C) 2005-2007 Cerion Armour-Brown <cerion@open-works.co.uk>
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -29,6 +29,8 @@
 */
 
 #include "pub_core_basics.h"
+#include "pub_core_vki.h"
+#include "pub_core_vkiscnums.h"
 #include "pub_core_threadstate.h"
 #include "pub_core_aspacemgr.h"
 #include "pub_core_debuglog.h"
@@ -50,8 +52,6 @@
 #include "priv_syswrap-linux.h"     /* for decls of linux-ish wrappers */
 #include "priv_syswrap-main.h"
 
-#include "vki_unistd.h"              /* for the __NR_* constants */
-
 
 /* ---------------------------------------------------------------------
    clone() handling
@@ -239,7 +239,7 @@ static SysRes do_clone ( ThreadId ptid,
    ThreadState* ctst = VG_(get_ThreadState)(ctid);
    ULong        word64;
    UWord*       stack;
-   NSegment*    seg;
+   NSegment const* seg;
    SysRes       res;
    vki_sigset_t blockall, savedmask;
 
@@ -313,6 +313,12 @@ static SysRes do_clone ( ThreadId ptid,
       ctst->client_stack_szB  = 0;
    }
 
+   /* Assume the clone will succeed, and tell any tool that wants to
+      know that this thread has come into existence.  If the clone
+      fails, we'll send out a ll_exit notification for it at the out:
+      label below, to clean up. */
+   VG_TRACK ( pre_thread_ll_create, ptid, ctid );
+
    if (flags & VKI_CLONE_SETTLS) {
       if (debug)
          VG_(printf)("clone child has SETTLS: tls at %p\n", child_tls);
@@ -344,6 +350,8 @@ static SysRes do_clone ( ThreadId ptid,
       /* clone failed */
       VG_(cleanup_thread)(&ctst->arch);
       ctst->status = VgTs_Empty;
+      /* oops.  Better tell the tool the thread exited in a hurry :-) */
+      VG_TRACK( pre_thread_ll_exit, ctid );
    }
 
    return res;
@@ -386,6 +394,7 @@ DECL_TEMPLATE(ppc32_linux, sys_mmap);
 DECL_TEMPLATE(ppc32_linux, sys_mmap2);
 DECL_TEMPLATE(ppc32_linux, sys_stat64);
 DECL_TEMPLATE(ppc32_linux, sys_lstat64);
+DECL_TEMPLATE(ppc32_linux, sys_fstatat64);
 DECL_TEMPLATE(ppc32_linux, sys_fstat64);
 DECL_TEMPLATE(ppc32_linux, sys_ipc);
 DECL_TEMPLATE(ppc32_linux, sys_clone);
@@ -672,7 +681,7 @@ PRE(sys_mmap2)
    SysRes r;
 
    // Exactly like old_mmap() except:
-   //  - the file offset is specified in pagesize units rather than bytes,
+   //  - the file offset is specified in 4K units rather than bytes,
    //    so that it can be used for files bigger than 2^32 bytes.
    PRINT("sys_mmap2 ( %p, %llu, %d, %d, %d, %d )",
          ARG1, (ULong)ARG2, ARG3, ARG4, ARG5, ARG6 );
@@ -682,7 +691,7 @@ PRE(sys_mmap2)
                  unsigned long, fd,    unsigned long, offset);
 
    r = ML_(generic_PRE_sys_mmap)( tid, ARG1, ARG2, ARG3, ARG4, ARG5, 
-                                       VKI_PAGE_SIZE * (Off64T)ARG6 );
+                                       4096 * (Off64T)ARG6 );
    SET_STATUS_from_SysRes(r);
 }
 
@@ -719,6 +728,20 @@ POST(sys_lstat64)
    }
 }
 
+PRE(sys_fstatat64)
+{
+   PRINT("sys_fstatat64 ( %d, %p(%s), %p )",ARG1,ARG2,ARG2,ARG3);
+   PRE_REG_READ3(long, "fstatat64",
+                 int, dfd, char *, file_name, struct stat64 *, buf);
+   PRE_MEM_RASCIIZ( "fstatat64(file_name)", ARG2 );
+   PRE_MEM_WRITE( "fstatat64(buf)", ARG3, sizeof(struct vki_stat64) );
+}
+
+POST(sys_fstatat64)
+{
+   POST_MEM_WRITE( ARG3, sizeof(struct vki_stat64) );
+}
+
 PRE(sys_fstat64)
 {
   PRINT("sys_fstat64 ( %d, %p )",ARG1,ARG2);
@@ -1009,8 +1032,11 @@ PRE(sys_clone)
 
 PRE(sys_sigreturn)
 {
+   /* See comments on PRE(sys_rt_sigreturn) in syswrap-amd64-linux.c for
+      an explanation of what follows. */
+
    ThreadState* tst;
-   PRINT("sigreturn ( )");
+   PRINT("sys_sigreturn ( )");
 
    vg_assert(VG_(is_valid_tid)(tid));
    vg_assert(tid >= 1 && tid < VG_N_THREADS);
@@ -1027,38 +1053,23 @@ PRE(sys_sigreturn)
    //ML_(fixup_guest_state_to_restart_syscall)(&tst->arch);
    // Should we do something equivalent on ppc32?  Who knows.
 
+   /* Restore register state from frame and remove it */
    VG_(sigframe_destroy)(tid, False);
 
-   /* For unclear reasons, it appears we need the syscall to return
-      without changing R3.  Since R3 is the return value, and can
-      denote either success or failure, we must set up so that the
-      driver logic copies it back unchanged.  Also, note R3 is of
-      the guest registers written by VG_(sigframe_destroy). */
-   /* jrs 16 Nov 05: for some reason this occasionally causes the 
-      is-this-a-sane-error-value sanity check to fail:
-      m_syswrap/syswrap-ppc32-linux.c:1037
-        (vgSysWrap_ppc32_linux_sys_sigreturn_before): 
-        Assertion 'wzz >= 0 && wzz < 10000' failed.
-      Hence use a sanity-check-free version.  
-      Perhaps we should ignore CR0.S0 here?
-      In general I have no idea what this is for or if it is necessary.
-      It's a conceptual copy-n-paste from the x86 equivalent, but I'm 
-      equally unclear as to whether it is needed there either.
-   */
-   SET_STATUS_from_SysRes_NO_SANITY_CHECK(
-      VG_(mk_SysRes_ppc32_linux)( 
-         tst->arch.vex.guest_GPR3,
-         /* get CR0.SO */
-         (LibVEX_GuestPPC32_get_CR( &tst->arch.vex ) >> 28) & 1
-      )
-   );
+   /* Tell the driver not to update the guest state with the "result",
+      and set a bogus result to keep it happy. */
+   *flags |= SfNoWriteResult;
+   SET_STATUS_Success(0);
 
-   /* Check to see if some any signals arose as a result of this. */
+   /* Check to see if any signals arose as a result of this. */
    *flags |= SfPollAfter;
 }
 
 PRE(sys_rt_sigreturn)
 {
+   /* See comments on PRE(sys_rt_sigreturn) in syswrap-amd64-linux.c for
+      an explanation of what follows. */
+
    ThreadState* tst;
    PRINT("rt_sigreturn ( )");
 
@@ -1077,18 +1088,15 @@ PRE(sys_rt_sigreturn)
    //ML_(fixup_guest_state_to_restart_syscall)(&tst->arch);
    // Should we do something equivalent on ppc32?  Who knows.
 
+   /* Restore register state from frame and remove it */
    VG_(sigframe_destroy)(tid, True);
 
-   /* See comments above in PRE(sys_sigreturn) about this. */
-   SET_STATUS_from_SysRes_NO_SANITY_CHECK(
-      VG_(mk_SysRes_ppc32_linux)( 
-         tst->arch.vex.guest_GPR3,
-         /* get CR0.SO */
-         (LibVEX_GuestPPC32_get_CR( &tst->arch.vex ) >> 28) & 1
-      )
-   );
+   /* Tell the driver not to update the guest state with the "result",
+      and set a bogus result to keep it happy. */
+   *flags |= SfNoWriteResult;
+   SET_STATUS_Success(0);
 
-   /* Check to see if some any signals arose as a result of this. */
+   /* Check to see if any signals arose as a result of this. */
    *flags |= SfPollAfter;
 }
 
@@ -1583,10 +1591,10 @@ const SyscallTableEntry ML_(syscall_table)[] = {
    GENXY(__NR_fstatfs,           sys_fstatfs),           // 100
 //..    LINX_(__NR_ioperm,            sys_ioperm),            // 101
    PLAXY(__NR_socketcall,        sys_socketcall),        // 102
-//..    LINXY(__NR_syslog,            sys_syslog),            // 103
+   LINXY(__NR_syslog,            sys_syslog),            // 103
    GENXY(__NR_setitimer,         sys_setitimer),         // 104
-//.. 
-//..    GENXY(__NR_getitimer,         sys_getitimer),         // 105
+
+   GENXY(__NR_getitimer,         sys_getitimer),         // 105
    GENXY(__NR_stat,              sys_newstat),           // 106
    GENXY(__NR_lstat,             sys_newlstat),          // 107
    GENXY(__NR_fstat,             sys_newfstat),          // 108
@@ -1599,7 +1607,7 @@ const SyscallTableEntry ML_(syscall_table)[] = {
    GENXY(__NR_wait4,             sys_wait4),             // 114
 //.. 
 //..    //   (__NR_swapoff,           sys_swapoff),           // 115 */Linux 
-//..    LINXY(__NR_sysinfo,           sys_sysinfo),           // 116
+   LINXY(__NR_sysinfo,           sys_sysinfo),           // 116
    PLAXY(__NR_ipc,               sys_ipc),               // 117
    GENX_(__NR_fsync,             sys_fsync),             // 118
    PLAX_(__NR_sigreturn,         sys_sigreturn),         // 119 ?/Linux
@@ -1668,7 +1676,7 @@ const SyscallTableEntry ML_(syscall_table)[] = {
 //.. 
    LINX_(__NR_setresgid,         sys_setresgid),         // 169
    LINXY(__NR_getresgid,         sys_getresgid),         // 170
-   LINX_(__NR_prctl,             sys_prctl),             // 171
+   LINXY(__NR_prctl,             sys_prctl),             // 171
    PLAX_(__NR_rt_sigreturn,      sys_rt_sigreturn),      // 172
    LINXY(__NR_rt_sigaction,      sys_rt_sigaction),      // 173
 
@@ -1745,7 +1753,7 @@ const SyscallTableEntry ML_(syscall_table)[] = {
 //.. 
    LINX_(__NR_set_tid_address,   sys_set_tid_address),   // 232
 
-//..    LINX_(__NR_fadvise64,         sys_fadvise64),         // 233 */(Linux?)
+   LINX_(__NR_fadvise64,         sys_fadvise64),         // 233 */(Linux?)
    LINX_(__NR_exit_group,        sys_exit_group),        // 234
 //..    GENXY(__NR_lookup_dcookie,    sys_lookup_dcookie),    // 235
    LINXY(__NR_epoll_create,      sys_epoll_create),      // 236
@@ -1798,6 +1806,22 @@ const SyscallTableEntry ML_(syscall_table)[] = {
    LINX_(__NR_inotify_init,  sys_inotify_init),               // 275
    LINX_(__NR_inotify_add_watch,  sys_inotify_add_watch),     // 276
    LINX_(__NR_inotify_rm_watch,   sys_inotify_rm_watch),      // 277
+
+   LINXY(__NR_openat,            sys_openat),            // 286
+   LINX_(__NR_mkdirat,           sys_mkdirat),           // 287
+   LINX_(__NR_mknodat,           sys_mknodat),           // 288
+   LINX_(__NR_fchownat,          sys_fchownat),          // 289
+   LINX_(__NR_futimesat,         sys_futimesat),         // 290
+   PLAXY(__NR_fstatat64,         sys_fstatat64),         // 291
+   LINX_(__NR_unlinkat,          sys_unlinkat),          // 292
+   LINX_(__NR_renameat,          sys_renameat),          // 293
+   LINX_(__NR_linkat,            sys_linkat),            // 294
+   LINX_(__NR_symlinkat,         sys_symlinkat),         // 295
+   LINX_(__NR_readlinkat,        sys_readlinkat),        // 296
+   LINX_(__NR_fchmodat,          sys_fchmodat),          // 297
+   LINX_(__NR_faccessat,         sys_faccessat),         // 298
+   LINX_(__NR_set_robust_list,   sys_set_robust_list),   // 299
+   LINXY(__NR_get_robust_list,   sys_get_robust_list),   // 300
 };
 
 const UInt ML_(syscall_table_size) = 
