@@ -39,24 +39,33 @@
 
 #include "pub_tool_machine.h"
 
-#if defined(VGA_x86)
+#if defined(VGP_x86_linux)
 #  define VG_ELF_DATA2XXX     ELFDATA2LSB
 #  define VG_ELF_MACHINE      EM_386
 #  define VG_ELF_CLASS        ELFCLASS32
-#elif defined(VGA_amd64)
+#  undef  VG_PLAT_USES_PPCTOC
+#elif defined(VGP_amd64_linux)
 #  define VG_ELF_DATA2XXX     ELFDATA2LSB
 #  define VG_ELF_MACHINE      EM_X86_64
 #  define VG_ELF_CLASS        ELFCLASS64
-#elif defined(VGA_ppc32)
+#  undef  VG_PLAT_USES_PPCTOC
+#elif defined(VGP_ppc32_linux)
 #  define VG_ELF_DATA2XXX     ELFDATA2MSB
 #  define VG_ELF_MACHINE      EM_PPC
 #  define VG_ELF_CLASS        ELFCLASS32
-#elif defined(VGA_ppc64)
+#  undef  VG_PLAT_USES_PPCTOC
+#elif defined(VGP_ppc64_linux)
 #  define VG_ELF_DATA2XXX     ELFDATA2MSB
 #  define VG_ELF_MACHINE      EM_PPC64
 #  define VG_ELF_CLASS        ELFCLASS64
+#  define VG_PLAT_USES_PPCTOC 1
+#elif defined(VGO_aix5)
+#  undef  VG_ELF_DATA2XXX
+#  undef  VG_ELF_MACHINE
+#  undef  VG_ELF_CLASS
+#  define VG_PLAT_USES_PPCTOC 1
 #else
-#  error Unknown arch
+#  error Unknown platform
 #endif
 
 #if defined(VGA_x86)
