@@ -1,7 +1,9 @@
 
 #include <signal.h>
 #include <stdio.h>
-#include <sys/syscall.h>
+#if !defined(_AIX)
+# include <sys/syscall.h>
+#endif
 #include <unistd.h>
 
 // Reg test for bug #93328: we were using too-big sigset types, and thus
@@ -9,12 +11,9 @@
 
 int main(void)
 {
-   int x[6], *s, *os, i, sysno;
+   int x[6], *s, *os, i;
 
-   sysno = __NR_rt_sigprocmask;
-#ifdef __NR_sigprocmask
-   sysno = __NR_sigprocmask;
-#endif
+#if defined(__NR_sigprocmask) && !defined(__powerpc64__) && !defined(_AIX)
 
    x[0] = 0x11111111;
    x[1] = 0x89abcdef;
@@ -30,7 +29,7 @@ int main(void)
    // blocked as perl has been known to leave some signals blocked
    // when starting child processes which can cause failures in
    // this test unless we reset things here.
-   syscall(sysno, SIG_SETMASK, os, NULL);
+   syscall(__NR_sigprocmask, SIG_SETMASK, os, NULL);
 
    fprintf(stderr, "before\n");
    for (i = 0; i < 6; i++) {
@@ -38,7 +37,7 @@ int main(void)
    }
    fprintf(stderr, "\n");
 
-   syscall(sysno, SIG_BLOCK, s, os);
+   syscall(__NR_sigprocmask, SIG_BLOCK, s, os);
 
    fprintf(stderr, "after1\n");
    for (i = 0; i < 6; i++) {
@@ -46,13 +45,19 @@ int main(void)
    }
    fprintf(stderr, "\n");
    
-   syscall(sysno, SIG_BLOCK, s, os);
+   syscall(__NR_sigprocmask, SIG_BLOCK, s, os);
 
    fprintf(stderr, "after2\n");
    for (i = 0; i < 6; i++) {
       fprintf(stderr, "%x ", x[i]);
    }
    fprintf(stderr, "\n");
-   
+
+#else
+
+   fprintf(stderr, "__NR_sigprocmask not supported on this platform\n");
+
+#endif
+
    return(0);
 }
